"""
Demonstrates using a `LoopStep` for iterative refinement.

The pipeline will generate text, then loop through a "review and edit"
cycle until the text meets a quality standard or the max number of
loops is reached. For more details, see docs/pipeline_looping.md.
"""
import asyncio
from typing import Any

from pydantic import BaseModel

from flujo import Flujo, Step, Pipeline


class FunctionAgent:
    """Wraps an async function so it can be used as a pipeline agent."""

    def __init__(self, func: Any) -> None:
        self.func = func

    async def run(self, data: Any = None, **kwargs: Any) -> Any:
        return await self.func(data, **kwargs)


class TextEdit(BaseModel):
    """A model to hold the text and feedback during the loop."""
    text: str
    feedback: str = "No feedback yet."
    is_good_enough: bool = False


# Agents for our iterative workflow
async def generate_text_agent(prompt: str) -> TextEdit:
    """Generates the initial draft."""
    print("✍️ Generating initial draft...")
    return TextEdit(text="Python is a language.")


async def edit_and_review_agent(draft: TextEdit) -> TextEdit:
    """Reviews text. If it's too short, it adds to it and provides feedback."""
    print(f"🧐 Reviewing draft: '{draft.text}'")
    if len(draft.text) < 30:
        print("   -> Draft is too short. Adding more detail.")
        draft.text += " It is known for its simple syntax."
        draft.feedback = "The text was too short, so I expanded it."
        draft.is_good_enough = False
    else:
        print("   -> Draft meets length requirement. Exiting loop.")
        draft.feedback = "The text looks good."
        draft.is_good_enough = True
    return draft


# The body of our loop is a single "edit and review" step.
loop_body_pipeline = Pipeline.from_step(
    Step("EditAndReview", agent=FunctionAgent(edit_and_review_agent))
)

# The `LoopStep` will run the `loop_body_pipeline` repeatedly.
loop_step = Step.loop_until(
    name="IterativeRefinementLoop",
    loop_body_pipeline=loop_body_pipeline,
    # The `exit_condition_callable` checks the output of the last step in the
    # loop body. It receives that output and the optional pipeline context.
    exit_condition_callable=lambda body_output, ctx: body_output.is_good_enough,
    max_loops=5,  # A safeguard to prevent infinite loops.
)

# The full pipeline: generate an initial version, then enter the refinement loop.
full_pipeline = Step(
    "GenerateInitialDraft", agent=FunctionAgent(generate_text_agent)
) >> loop_step

runner = Flujo(full_pipeline)


async def main() -> None:
    print("🚀 Starting iterative refinement pipeline...\n")
    result = None
    async for item in runner.run_async("Write a sentence about Python."):
        result = item

    print("\n✅ Pipeline finished!")
    final_result = result.step_history[-1]

    print(f"\nFinal Output: '{final_result.output.text}'")
    print(f"Final Feedback: '{final_result.output.feedback}'")
    print(f"Iterations (attempts): {final_result.attempts}")


if __name__ == "__main__":
    asyncio.run(main())
