import asyncio
import pytest
from pydantic import BaseModel

from flujo.application.flujo_engine import Flujo
from flujo.domain import Step
from flujo.testing.utils import StubAgent, gather_result
from flujo.domain.models import PipelineResult


class Ctx(BaseModel):
    count: int = 0


class AddOneAgent:
    async def run(self, data: int, *, pipeline_context: Ctx | None = None) -> int:
        if pipeline_context:
            pipeline_context.count += 1
        return data + 1


@pytest.mark.asyncio
async def test_pipeline_runner_shared_context_flow() -> None:
    step1 = Step("a", AddOneAgent())
    step2 = Step("b", AddOneAgent())
    runner = Flujo(step1 >> step2, context_model=Ctx, initial_context_data={"count": 0})
    result = await gather_result(runner, 1)
    assert result.final_pipeline_context.count == 2
    assert result.step_history[-1].output == 3


@pytest.mark.asyncio
async def test_existing_agents_without_context() -> None:
    agent = StubAgent(["ok"])
    step = Step("s", agent)
    runner = Flujo(step)
    result = await gather_result(runner, "hi")
    assert result.step_history[0].output == "ok"


class TestContext(BaseModel):
    counter: int = 0


class IncrementAgent:
    async def run(
        self, data: str, *, pipeline_context: TestContext | None = None
    ) -> str:
        if pipeline_context:
            pipeline_context.counter += 1
        return data


@pytest.mark.asyncio
async def test_concurrent_runs_with_typed_context_are_isolated() -> None:
    step = Step("inc", IncrementAgent())
    runner = Flujo(step, context_model=TestContext)

    async def run_one() -> PipelineResult:
        return await gather_result(runner, "input")

    result1, result2 = await asyncio.gather(run_one(), run_one())

    assert result1.final_pipeline_context.counter == 1
    assert result2.final_pipeline_context.counter == 1
